//  the lone titlebar in fullscreen-ish mode is bad..
//  better just squish to left... and no title - bar or more subdued titlebar?
//

#include "local.conf"
#if CTX_VT

#if !__COSMOPOLITAN__
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <math.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#endif

#include "ctx.h"
#include "terminal-keyboard.h"

#include "ui/ui.h"
#include "ui/ui.c"
#undef em

static Ui *ui = NULL;

  CtxClient *second = NULL; // XXX at least give better name
static float animation_duration = 0.175f;//25f;
static Ctx *ctx = NULL; // initialized in main
static char *ctx_config = NULL;
static uint64_t ctx_config_timestamp = 0;
static float ctx_cfg_dirty = 0.0;
static float help_elapsed = 0.0f;

static int force_full_size = 0;
static int experimental = 0;
static float backlight = 100.0f;
static float volume = 0.2f;

static float full_time = 0.0f;
float dock_dim = 4.0f;

//static int cfg_has_reloaded = 0;
static int layout_no = 2;

#define ENABLE_ROTATE 0

#include "cfg.h"


static int locked = 0;

void ctx_set_focus_cb (Ctx *ctx, void(*focus_cb)(Ctx *ctx, int id, void *user_data), void *user_data);

typedef struct _CtxClient CtxClient;
CtxList *ctx_clients(Ctx *ctx);

int ctx_client_get_width (Ctx *ctx, int id);
int ctx_client_get_height (Ctx *ctx, int id);
int ctx_client_get_x (Ctx *ctx, int id);
int ctx_client_get_y (Ctx *ctx, int id);

void ctx_screenshot (Ctx *ctx, const char *path);
void
vt_screenshot (const char *output_path)
{
  ctx_screenshot (ctx, output_path);
}

void ctx_client_lock (CtxClient *client);
void ctx_client_unlock (CtxClient *client);

#define VT_RECORD 0
static int terminal_no_new_tab = 0;

static char *execute_self = NULL;

static float font_size    = 19.7f;
static float line_spacing = 1.0f;
static float baseline     = 0.78f;


float add_x = 10;
float add_y = 100;

/********************/

static char *ctx_term_config_path (void)
{
  static char *res = NULL; // will be leaked
  if (res) return res;
  res = ctx_strdup_printf ("%s/.config/ctx/ctx.ini", getenv("HOME"));
  return res;
}

static const char *ctx_find_shell_command (void)
{
  if (access ("/.flatpak-info", F_OK) != -1)
  {
    static char ret[512];
    char buf[256];
    FILE *fp = popen("flatpak-spawn --host getent passwd $USER|cut -f 7 -d :", "r");
    if (fp)
    {
      while (fgets (buf, sizeof(buf), fp) != NULL)
      {
        if (buf[strlen(buf)-1]=='\n')
          buf[strlen(buf)-1]=0;
        sprintf (ret, "flatpak-spawn --env=TERM=xterm --host %s", buf);
      }
      pclose (fp);
      return ret;
    }
  }

  if (getenv ("SHELL"))
  {
    return getenv ("SHELL");
  }
  int i;
  const char *command = NULL;
  struct stat stat_buf;
  static const char *alts[][2] =
  {
    {"/bin/bash",     "/bin/bash"},
    {"/usr/bin/bash", "/usr/bin/bash"},
    {"/bin/sh",       "/bin/sh"},
    {"/usr/bin/sh",   "/usr/bin/sh"},
    {NULL, NULL}
  };
  for (i = 0; alts[i][0] && !command; i++)
    {
      lstat (alts[i][0], &stat_buf);
      if (S_ISREG (stat_buf.st_mode) || S_ISLNK (stat_buf.st_mode) )
        { command = alts[i][1]; }
    }
  return command;
}
static int in_overview = 0;

/*****************/
int is_big = 2;

#define flag_is_set(a, f) (((a) & (f))!=0)
#define flag_set(a, f)    ((a) |= (f));
#define flag_unset(a, f)  ((a) &= ~(f));

int add_tab (Ctx  *ctx, const char *commandline, int can_launch)
{
  float titlebar_h = ctx_height (ctx)/40;
  int was_maximized = ctx_client_is_maximized (ctx, ctx_clients_active (ctx));
  int flags = CTX_CLIENT_UI_RESIZABLE |  CTX_CLIENT_TITLEBAR;
  if (can_launch) flags |= CTX_CLIENT_CAN_LAUNCH;

  CtxClient *active = ctx_client_new (ctx, commandline, add_x, add_y,
    ctx_width(ctx)/2, (ctx_height (ctx) - titlebar_h)/2,
    font_size,
    flags, NULL, NULL);

  VT *vt = ctx_client_vt (active);
  vt_set_line_spacing (vt, line_spacing);
  vt_set_baseline (vt, baseline);
  add_y += ctx_height (ctx) / 20;
  add_x += ctx_height (ctx) / 20;

  if (was_maximized)
  {
    ctx_client_maximize (ctx, ctx_client_id (active));
  }

  if (add_y + ctx_height(ctx)/2 > ctx_client_max_y_pos (ctx))
  {
    add_y = ctx_client_min_y_pos (ctx);
    add_x -= ctx_height (ctx) / 40 * 4;
  }
  ctx_queue_draw (ctx);
  ctx_client_focus (ctx, ctx_client_id (active));

  if (layout_no == 2)
    ctx_client_raise_almost_top (ctx, ctx_client_id (active));
  else
    ctx_client_raise_top (ctx, ctx_client_id (active));
  return ctx_client_id (active);
}

int add_tab_argv (Ctx  *ctx, char **argv, int can_launch)
{
  float titlebar_h = ctx_height (ctx)/40;
  int was_maximized = ctx_client_is_maximized (ctx, ctx_clients_active (ctx));
  int flags = CTX_CLIENT_UI_RESIZABLE |  CTX_CLIENT_TITLEBAR;
  if (can_launch) flags |= CTX_CLIENT_CAN_LAUNCH;

  CtxClient *active = ctx_client_new_argv (ctx, argv, add_x, add_y,
                    ctx_width(ctx)/2, (ctx_height (ctx) - titlebar_h)/2,
                    font_size,
                    flags, NULL, NULL);
  add_y += ctx_height (ctx) / 20;
  add_x += ctx_height (ctx) / 20;

  if (was_maximized)
  {
    ctx_client_maximize (ctx, ctx_client_id (active));
  }

  if (add_y + ctx_height(ctx)/2 > ctx_client_max_y_pos (ctx))
  {
    add_y = ctx_client_min_y_pos (ctx);
    add_x -= ctx_height (ctx) / 40 * 4;
  }
  ctx_queue_draw (ctx);
  return ctx_client_id (active);
}

static void remove_tab_cb (CtxEvent *event, void *data, void *data2)
{
  event->stop_propagate = 1;
  Ctx *ctx = event->ctx;
  int active_id = ctx_clients_active (ctx);
  CtxClient *active = active_id>=0?ctx_client_by_id (ctx, active_id):NULL;
  if (active)
     ctx_client_remove (ctx, active);
  ctx_queue_draw (ctx);
}

static void add_tab_cb (CtxEvent *event, void *data, void *data2)
{
  event->stop_propagate = 1;
  if (!terminal_no_new_tab)
    add_tab (event->ctx, ctx_find_shell_command(), 1);
}

int ctx_clients_tab_to_id (Ctx *ctx, int tab_no);

void switch_to_tab (Ctx *ctx, int tab_no)
{
  if (layout_no == 2)
  {
    CtxList *l = ctx_clients (ctx);
    int client_count = ctx_list_length (l);

    if (tab_no == 0)
      tab_no = 1;
    else if (tab_no == 1)
      tab_no = 0;

    CtxClient *client = ctx_list_nth_data (l, client_count - 1 - tab_no);
    int id = ctx_client_id (client);
    if (tab_no <= 1)
    {
      if (client)
        ctx_client_focus (ctx, id);
    }
    else
    {
      if (client)
      {
        ctx_client_focus (ctx, id);
        ctx_client_raise_almost_top (ctx, id);
      }
    }
    ctx_queue_draw (ctx);
    return;
  }


  int id = ctx_clients_tab_to_id (ctx, tab_no);
  if (id >= 0)
  {
    ctx_client_maximize (ctx, id);
  }
}

CtxClient *ctx_client_by_id (Ctx *ctx, int id);

CtxEvent *ctx_event_copy (CtxEvent *event);



int         vt_get_scroll           (VT *vt);
void        vt_set_scroll           (VT *vt, int scroll);

#if ENABLE_ROTATE
static float global_rotation = 0.0f;

static void rotate_cb (CtxEvent *event, void *a, void *b)
{
  global_rotation += M_PI / 2;
  ctx_queue_draw (event->ctx);
}
#endif

void vt_set_font_size (VT *vt, float font_size);
void vt_recompute_cols (VT *vt);

void set_new_font_size (Ctx *ctx, float new_font_size)
{
  font_size = new_font_size;
  cfg_set_f32 (&ctx_config, "ctx", "font_size", new_font_size, NULL);
  for (CtxList *l = ctx_clients (ctx); l; l = l->next)
  {
    CtxClient *client = l->data;
    VT *vt = ctx_client_vt (client);
    vt_set_font_size (vt, new_font_size); //
    vt_set_line_spacing (vt, line_spacing); //
    vt_set_baseline (vt, baseline); //
    vt_recompute_cols (vt);
  }
  ctx_queue_draw (ctx);
}

static void wm_focus_cb (Ctx *ctx, int focused_id, void *user_data)
{
  ctx_client_raise_top (ctx, focused_id);
}

static void raise_to_top (CtxEvent *event, void *data1, void *data2)
{
  CtxClient *client = data1;
  Ctx *ctx = event->ctx;
  ctx_client_raise_top (ctx, ctx_client_id (client));
  ctx_queue_draw (ctx);
  ctx_client_focus (ctx, ctx_client_id (client));
  event->stop_propagate = 1;
}

static void overview_event (CtxEvent *event, void *a, void *b);

static void terminal_handle_event (Ctx        *ctx,
                          CtxEvent   *ctx_event,
                          const char *event)
{
  int active_id = ctx_clients_active (ctx);
  CtxClient *active = active_id>=0?ctx_client_by_id (ctx, active_id):NULL;
  if (!active)
  {
     printf("no active client\n");
    return;
  }
  //if (active->internal)
  //  return;
  VT *vt = ctx_client_vt (active);

  CtxClient *client = active; //vt_get_client (vt);

  if (!vt)
  {
     ctx_client_add_event (active, ctx_event);
     return;
  }

  ctx_client_lock (client);

  if (locked)
  {
    ctx_client_feed_keystring (active, ctx_event, event);
    ctx_client_unlock (client);
    return;
  }
  
  if (!strcmp (event, "F11") ||
      !strcmp (event, "shift-control-f"))
  {
    int val = !ctx_get_fullscreen (ctx);
    cfg_set_boolean (&ctx_config, "ctx", "fullscreen", val, NULL);
    ctx_set_fullscreen (ctx, val);
  }
  else if (!strcmp (event, "shift-return"))
  {
    ctx_client_feed_keystring (active, ctx_event, "return");
  }
#if ENABLE_ROTATE
  else if (!strcmp (event, "shift-control-r") )
  {
     CtxEvent dummy;dummy.ctx=ctx;
     rotate_cb (&dummy, NULL, NULL);
  }
#endif
  else if (!strcmp (event, "shift-control-v") )
  {
    char *text = ctx_get_clipboard (ctx);
    if (text)
      {
        ctx_client_paste (ctx, active_id, text);
        free (text);
      }
  }
  else if (!strcmp (event, "shift-control-c") && vt)
  {
    char *text = ctx_client_get_selection (ctx, active_id);
    if (text)
      {
        ctx_set_clipboard (ctx, text);
        free (text);
      }
  }
  else if (!strcmp (event, "shift-control-t"))
  {
  if (!terminal_no_new_tab)
    add_tab (ctx, ctx_find_shell_command(), 1);
  }
  else if (!strcmp (event, "shift-control-o"))
  {
   CtxEvent dummy;dummy.ctx=ctx;
   overview_event (&dummy, NULL, NULL);
  }
  else if (!strcmp (event, "shift-control-n") )
  {
    if (!terminal_no_new_tab)
    {
             // XXX : we should also avoid this when running with kms/fb !
    pid_t pid;
    if ( (pid=fork() ) ==0)
      { 
        unsetenv ("CTX_VERSION");
        unsetenv ("CTX_BACKEND");
        execlp (execute_self, execute_self, NULL);
        exit (0);
      }
    }
  }
  else if (!strcmp (event, "alt-1"))   switch_to_tab(ctx, 0);
  else if (!strcmp (event, "alt-2"))   switch_to_tab(ctx, 1);
  else if (!strcmp (event, "alt-3"))   switch_to_tab(ctx, 2);
  else if (!strcmp (event, "alt-4"))   switch_to_tab(ctx, 3);
  else if (!strcmp (event, "alt-5"))   switch_to_tab(ctx, 4);
  else if (!strcmp (event, "alt-6"))   switch_to_tab(ctx, 5);
  else if (!strcmp (event, "alt-7"))   switch_to_tab(ctx, 6);
  else if (!strcmp (event, "alt-8"))   switch_to_tab(ctx, 7);
  else if (!strcmp (event, "alt-9"))   switch_to_tab(ctx, 8);
  else if (!strcmp (event, "alt-0"))   switch_to_tab(ctx, 9);
  else if (!strcmp (event, "shift-control-q") )
  {
    ctx_exit (ctx);
  }
  else if (!strcmp (event, "control-tab") ||
           !strcmp (event, "alt-tab"))
  {
     CtxList *l = ctx_clients (ctx);
     int found = 0;
     CtxClient *next = NULL;
     for (; l; l = l->next)
     {
       CtxClient *client = l->data;
       if (found)
       {
         if (!next) next = client;
       }
       if (client == active) found = 1;
     }
     if (!next)
       next = ctx_clients (ctx)->data;
     ctx_client_focus (ctx, ctx_client_id (next));
     ctx_queue_draw (ctx);
  }
  else if (!strcmp (event, "shift-control-tab") ||
           !strcmp (event, "shift-alt-tab"))
  {
     CtxList *l = ctx_clients (ctx);
     CtxClient *prev = NULL;
     if (l->data == active)
     {
       for (; l; l = l->next)
         prev = l->data;
     }
     else
     {
       for (; l; l = l->next)
       {
         CtxClient *client = l->data;
         if (client == active)
           break;
         prev = client;
       }
     }
     ctx_client_focus (ctx, ctx_client_id (prev));
     ctx_queue_draw (ctx);
  }
  else if (!strcmp (event, "shift-control-up") )
  {
    if (vt)
    {
      vt_set_scroll (vt, vt_get_scroll (vt)+1);
      ctx_queue_draw (ctx);
    }
  }
  else if (!strcmp (event, "shift-control-down") )
  {
    if (vt)
    {
      vt_set_scroll (vt, vt_get_scroll (vt)-1);
      ctx_queue_draw (ctx);
    }
  }
  else if (!strcmp (event, "shift-control-w") )
  {
    ctx_client_unlock (client);
    ctx_client_remove (ctx, client);
    ctx_queue_draw (ctx);
    return;
  }
  else if (  !strcmp (event, "control-+") 
           ||!strcmp (event, "shift-control-+") 
        )
  {
  }
  else if (( !strcmp (event, "control-=") ||!strcmp (event, "shift-control-=")))
  {
    int old_rows = ctx_height (ctx) / font_size;
    float new_font_size = ctx_height (ctx) * 1.0f / (old_rows-1.0f);
    float new_rows = ctx_height (ctx) / new_font_size;

    if (old_rows <=4) goto done;
    //fprintf (stderr, "+(%s)", event);
    while (fmodf(new_rows, 1.0f) > 0.2)
    {
      new_font_size /= 1.001;
      new_rows = ctx_height (ctx) / new_font_size;
    }

    set_new_font_size (ctx, new_font_size);
  }
  else if (!strcmp (event, "shift-control-m"))
  {
    is_big++;
    if (is_big>2)
      is_big = 0;
    ctx_queue_draw (ctx);
  }
  else if (!strcmp (event, "shift-control-e"))
  {
    layout_no++;
    if (layout_no>2)
      layout_no = 0;

    switch (layout_no)
    {
      case 0:
         ctx_set_focus_cb (ctx, NULL, NULL);
         for (CtxList *l = ctx_clients (ctx); l; l = l->next)
         {
           CtxClient *client = l->data;
           int id = ctx_client_id (client);
           ctx_client_maximize (ctx, id);
         }
        break;
      case 1:
         ctx_set_focus_cb (ctx, wm_focus_cb, NULL);
         for (CtxList *l = ctx_clients (ctx); l; l = l->next)
         {
           CtxClient *client = l->data;
           
           int id = ctx_client_id (client);
           ctx_client_unmaximize (ctx, id);
           int w = ctx_width (ctx)/2  - font_size;
           int h = ctx_height (ctx)/2 - font_size * 2;
           ctx_client_resize (ctx, id, w, h);
         }
        break;
      case 2:
         ctx_set_focus_cb (ctx, NULL, NULL);
         for (CtxList *l = ctx_clients (ctx); l; l = l->next)
         {
           CtxClient *client = l->data;
           
           int id = ctx_client_id (client);
           ctx_client_unmaximize (ctx, id);
           int w = ctx_width (ctx)/2  - font_size;
           int h = ctx_height (ctx)/2 - font_size * 2;
           ctx_client_resize (ctx, id, w, h);
         }
        break;
    }

    ctx_queue_draw (ctx);
  }
  else if (!strcmp (event, "control--"))
  {
    int old_rows = roundf(ctx_height (ctx) / font_size);
    float new_font_size = ctx_height (ctx) * 1.0f / (old_rows+1.0f);
    float new_rows = ctx_height (ctx) / new_font_size;
    //fprintf (stderr, "-(%s)", event);
    while (floorf(new_rows)==old_rows)//fmodf(new_rows, 1.0f) > 0.05)
    {
      new_font_size /= 1.001;
      new_rows = ctx_height (ctx) / new_font_size;
    }

    if (font_size < 8) font_size = 8;
    set_new_font_size (ctx, new_font_size);
  }
  else if (!strcmp (event, "shift-control-/") )
  {
    help_elapsed = 0.0f;
    ctx_queue_draw (ctx);
  }
  else if (!strcmp (event, "shift-control-j") )
  {
    CtxEvent dummy = {};
    dummy.ctx = ctx;
    raise_to_top (&dummy, second, NULL);
    //
  }
  else if (!strcmp (event, "shift-control-s") )
  {
    if (vt)
    {
      char *sel = ctx_client_get_selection (ctx, ctx_client_id (active));
      if (sel)
      {
        ctx_client_feed_keystring (active, ctx_event, sel);
        free (sel);
      }
    }
  }
  else if (!strncmp (event, "shift-control-", 14))
  {
  }
  else if (!strncmp (event, "alt-control-", 12))
  {
  }
  else if (!strncmp (event, "shift-alt-control-", 18))
  {
  }
  else
  {
    ctx_client_feed_keystring (active, ctx_event, event);
  }
  done:
  ctx_client_unlock (client);
}

static float consume_float (char **argv, int *i)
{
 float ret = 0.0;
 if (!argv[*i+1])
 {
   fprintf (stderr, "error parsing arguments\n");
   exit (2);
 }
 ret = atof (argv[*i+1]);
 (*i)++;
  int j;
  for (j = *i; argv[j]; j++)
  {
    argv[j-2] = argv[j];
  }
  argv[j-2]=0;
  (*i)-=2;
  return ret;
}
// good font size for 80col:
//
// 4 6 8 10 12 14 16 18 20 2
extern int on_screen_keyboard;
void ctx_osk_draw (Ctx *ctx);

#if GNU_C
static int malloc_trim_cb (Ctx *ctx, void *data)
{
  malloc_trim (64*1024);
  return 1;
}
#endif

int saved_cursor = -1;

static void terminal_show_cursor (CtxEvent *event, void *userdata, void *userdata2)
{
  if (saved_cursor == -1) return;
  ctx_set_cursor (event->ctx, CTX_CURSOR_ARROW);
  saved_cursor = -1;
}

static void terminal_hide_cursor (CtxEvent *event, void *userdata, void *userdata2)
{
  if (saved_cursor == -1)
  {
    saved_cursor = ctx_get_cursor (event->ctx);
    ctx_set_cursor (event->ctx, CTX_CURSOR_NONE);
  }

}


static void terminal_key_any (CtxEvent *event, void *userdata, void *userdata2)
{
  if (event->type == CTX_KEY_PRESS &&
      event->string &&
      !strcmp (event->string, "resize-event"))
  {
    Ctx *ctx = event->ctx;
    cfg_set_int (&ctx_config, "ctx", "width", ctx_width (ctx), NULL);
    cfg_set_int (&ctx_config, "ctx", "height", ctx_height (ctx), NULL);
    ctx_queue_draw (ctx);
  }
  else
  {
    switch (event->type)
    {
      case CTX_KEY_PRESS:
        terminal_handle_event (ctx, event, event->string);
        if (cfg_boolean (&ctx_config, "", "auto_hide_mouse"))
          terminal_hide_cursor (event, userdata, userdata2);
        break;
      case CTX_KEY_UP:
        { char buf[1024];
          snprintf (buf, sizeof(buf)-1, "keyup %i %i", event->unicode, event->state);
          terminal_handle_event (ctx, event, buf);
        }
        break;
      case CTX_KEY_DOWN:
        { char buf[1024];
          snprintf (buf, sizeof(buf)-1, "keydown %i %i", event->unicode, event->state);
          terminal_handle_event (ctx, event, buf);
        }
        break;
      default:
        break;
    }
  }
}

extern int _ctx_enable_hash_cache;

void ctx_client_titlebar_draw (Ctx *ctx, CtxClient *client,
                               float x, float y, float width, float titlebar_height);

static void icon_remove_tab (Ctx *ctx, float x, float y, float w, float h)
{
  ctx_save (ctx);
  ctx_translate (ctx, (x + 0.5 * w), (y + 0.5 * w));
  ctx_rotate (ctx, M_PI/4);
  ctx_translate (ctx, -0.5 * w, -0.5 * w);
  ctx_rectangle (ctx, 0.4 * w, 0.1 * h, 0.2 * w, 0.8 * h);
  ctx_rectangle (ctx, 0.1 * w, 0.4 * h, 0.8 * w, 0.2 * h);
  ctx_restore (ctx);
}

static void icon (Ctx *ctx, float x, float y, float w, float h, const char *name,
                  const char *drawing)
{
  ctx_save (ctx);
  ctx_text_align (ctx, CTX_TEXT_ALIGN_START);
  ctx_translate (ctx, x, y);
  ctx_scale (ctx, w, h);
  char *str = cfg_string_full (&ctx_config, "icons", name, drawing, NULL);
  ctx_parse (ctx, str);
  free (str);
  ctx_restore (ctx);
}

static void lock_cb (CtxEvent *event, void *a, void *b)
{
  system ("touch /tmp/ctx.lock");
  ctx_queue_draw (event->ctx);
}



int ctx_client_is_active_tab (Ctx *ctx, CtxClient *client);
void draw_panel (Ctx *ctx)
{
  float em = font_size;
  float tab_width = ctx_width (ctx) - 4 * em;

  ctx_save (ctx);

  ctx_reset_path (ctx);

  int tabs = 0;
  for (CtxList *l = ctx_clients (ctx); l; l = l->next)
  {
    CtxClient *client = l->data;
    if (ctx_client_flags (client) & CTX_CLIENT_MAXIMIZED)
    {
      tabs ++;
    }
  }

  if (tabs)
  tab_width /= tabs;

  ctx_reset_path (ctx);
  float x = 0.0f;

  for (CtxList *l = ctx_clients (ctx); l; l = l->next)
  {
    CtxClient *client = l->data;
    if (ctx_client_flags (client) & CTX_CLIENT_MAXIMIZED)
    {
      ctx_reset_path (ctx);
      if (ctx_client_is_active_tab (ctx, client))
      {
        ctx_rectangle (ctx, x + 0.2 * em, 0.0 * em,
                       tab_width - 0.4 * em, 0.15 * em);
        ctx_rgba (ctx, 1,1,1,0.5);
        ctx_fill (ctx);
      }
    }
    x += tab_width;
  }
  ctx_restore (ctx);
}

static char *set_title = NULL;
void vt_audio_task (VT *vt, int click);

void
terminal_update_title (const char *title)
{
  if (!title)
    title="NULL";
  if (set_title)
  {
    if (!strcmp (set_title, title))
    {
      return;
    }
    free (set_title);
  }
  set_title = strdup (title);
  ctx_windowtitle (ctx, set_title);
}

int ctx_input_pending (Ctx *ctx, int timeout);

void terminal_long_tap (CtxEvent *event, void *a, void *b)
{
  Ctx *ctx = event->ctx;
  on_screen_keyboard = !on_screen_keyboard;
  ctx_queue_draw (ctx);
}

static uint64_t path_modtime (const char *path)
{
  struct stat attrib;
  stat(path, &attrib);
  return attrib.st_mtime;
}

int commandline_argv_start = 0;

void ctx_client_draw (Ctx *ctx, CtxClient *client, float x, float y);

typedef struct OverviewPos {
  int id;
  CtxClient *client;
  float x0;
  float y0;
  float w0;
  float h0;
  float x1;
  float y1;
  float scale0;  /// is always 1.0f
  float scale1;

} OverviewPos;

static OverviewPos  *opos           = NULL;
static int           opos_count = 0;

static void overview_cleanup (Ctx *ctx)
{
  if (opos)
  {
    ctx_free (opos);
    opos = NULL;
  }
  opos_count = 0;
}

static void overview_init (Ctx *ctx)
{
  if (opos)
  {
    //return;
    overview_cleanup (ctx);
  }

  CtxList *clients = ctx_clients (ctx);
  float em = ctx_get_font_size (ctx);
  int n_clients         = ctx_list_length (clients);

  float gap = cfg_cached_f32 (&ctx_config, "style", "overview_gap");
  float screen_width = ctx_width (ctx) - cfg_cached_f32 (&ctx_config, "style", "dock_width") * em - gap;
  float screen_height = ctx_height (ctx);

  opos_count = n_clients;
  opos = (OverviewPos*)ctx_calloc(opos_count, sizeof(OverviewPos));

  int active_id = ctx_clients_active (ctx);
  int active_no = 0;
  CtxClient *active = NULL;

  int top_of_class = 0;
  {
    int rows = 1;
    int cols = 1;

    if (n_clients <= 4) { rows = 2; cols = 2;}

    while ( n_clients > cols * rows)
    {
      if (cols > rows * 1.2)
      {
         rows++;
      } else cols++;
    }

    float col_width = screen_width / cols;
    float row_height = screen_height / rows;

    float icon_width = col_width - gap;
    //float icon_height = row_height - em;
    
    int i = 0;

    for (CtxList *l = clients; l; l = l->next)
    {
      CtxClient *client = l->data;
      int client_id = ctx_client_id (client);

      int col = i % cols;
      int row = i / cols;

      opos[i].client = client;
      opos[i].id = client_id;
      opos[i].x0 = ctx_client_get_x (ctx, client_id);
      opos[i].y0 = ctx_client_get_y (ctx, client_id);
      opos[i].w0 = ctx_client_get_width (ctx, client_id);
      opos[i].h0 = ctx_client_get_height (ctx, client_id);
      opos[i].x1 = col * col_width + 0.5f * em;
      opos[i].y1 = row * row_height + 0.5f * em;

      opos[i].scale0 = 1.0f;
      opos[i].scale1 = icon_width/opos[i].w0;
    
      if ((ctx_client_flags (client) & CTX_CLIENT_MAXIMIZED))
      {
         if (!ctx_client_is_active_tab (ctx, client))
         {
           opos[i].x0 = opos[i].x1;
           opos[i].y0 = opos[i].y1;
           opos[i].scale0 = opos[i].scale1;
         }
         top_of_class = i;
      }
      if (client_id == active_id){
              active_no = i;
              active = client;
      }

      i++;
    }
  }


  if (!(ctx_client_flags (active) & CTX_CLIENT_MAXIMIZED))
    top_of_class = n_clients - 1;


  if (active_no != top_of_class)
  {
    OverviewPos temp = opos[top_of_class];
    opos[top_of_class] = opos[active_no];
    opos[active_no] = temp;
  }
}



#define CTX_TERM_UNLOCK_SH   \
        "sh -c \"echo 'ctx is locked'; while [ -f /tmp/ctx.lock ];do su `whoami` -c 'rm /tmp/ctx.lock';done\""


void ctx_client_use_images (Ctx *ctx, CtxClient *client);

void ctx_term_lock_screen (Ctx *ctx)
{
  //ctx_rgb (ctx, 1,0.0,0);
  //ctx_paint (ctx);
  ctx_rgb (ctx, 1,1,1);

  CtxList *clients = ctx_clients (ctx);

  CtxClient *last = NULL;

  for (CtxList *l = clients; l; l = l->next)
  {
    CtxClient *client = l->data;
    if (l->next)
      ctx_client_use_images (ctx, client);
    last = client;
  }

  if (last)
    ctx_client_draw (ctx, last, 0, 0);
}

static int clients_draw (Ctx *ctx)
{
  CtxList *clients = ctx_clients (ctx);
  int n_clients = ctx_list_length (clients);

  if (n_clients == 1) {
    CtxClient *client = clients->data;
    int flags = ctx_client_flags (client);
    if (client && flag_is_set(flags, CTX_CLIENT_MAXIMIZED))
    {
      ctx_client_draw (ctx, client, 0, 0);
      return 0;
    }
  }
  //float em = ctx_get_font_size (ctx);
#if 0
  float screen_width = ctx_width (ctx) - 3 * em;
  float screen_height = ctx_height (ctx);
#endif
  int active_id = ctx_clients_active (ctx);
  CtxClient *active = active_id>=0?ctx_client_by_id (ctx, active_id):NULL;

  for (CtxList *l = clients; l; l = l->next)
  {
    CtxClient *client = l->data;
    int flags = ctx_client_flags (client);
    if (flag_is_set(flags, CTX_CLIENT_MAXIMIZED))
    {
       if (client == active)
        ctx_client_draw (ctx, client, 0, 0);
      else
        ctx_client_use_images (ctx, client);
    }
  }

  return 0;
}

static void load_palette(void)
{
  uint8_t rgba[4];
  cfg_rgba_full (&ctx_config, "palette", "color15", rgba, "#ffffff", NULL);
  vt_set_palette(15, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color14", rgba, "#1de1ed", NULL);
  vt_set_palette(14, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color13", rgba, "#d66fed", NULL);
  vt_set_palette(13, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color12", rgba, "#82b4ed", NULL);
  vt_set_palette(12, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color11", rgba, "#e9d808", NULL);
  vt_set_palette(11, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color10", rgba, "#99edba", NULL);
  vt_set_palette(10, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color9", rgba, "#edac82", NULL);
  vt_set_palette(9, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color8", rgba, "#6f6f6f", NULL);
  vt_set_palette(8, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color7", rgba, "#c3c3c3", NULL);
  vt_set_palette(7, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color6", rgba, "#3b6bb1", NULL);
  vt_set_palette(6, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color5", rgba, "#ab4adf", NULL);
  vt_set_palette(5, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color4", rgba, "#2424ed", NULL);
  vt_set_palette(4, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color3", rgba, "#878453", NULL);
  vt_set_palette(3, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color2", rgba, "#4aa08b", NULL);
  vt_set_palette(2, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color1", rgba, "#a02929", NULL);
  vt_set_palette(1, rgba[0], rgba[1], rgba[2]);
  cfg_rgba_full (&ctx_config, "palette", "color0", rgba, "#000000", NULL);
  vt_set_palette(0, rgba[0], rgba[1], rgba[2]);
}

static void load_fonts(Ctx *ctx)
{
#if CTX_HARFBUZZ
  #if 0
      cfg_set_string (&ctx_config, "fonts", "Mono", "/usr/share/fonts/monaspace/MonaspaceNeon-Regular.otf", NULL);
      cfg_set_string (&ctx_config, "fonts", "Mono Bold", "/usr/share/fonts/monaspace/MonaspaceNeon-Bold.otf", NULL);
      cfg_set_string (&ctx_config, "fonts", "Mono Italic", "/usr/share/fonts/monaspace/MonaspaceNeon-Italic.otf", NULL);
      cfg_set_string (&ctx_config, "fonts", "Regular", "/usr/share/fonts/cantarell/Cantarell-VF.otf", NULL);
  #endif

  char *fonts[]={ "Mono Bold Italic", 
                  "Mono Italic",
                  "Mono Bold",
                  "Mono",
                  "Regular",
                  NULL, };
  for (int i = 0; fonts[i]; i++)
  {
    char *font_path = cfg_string_full (&ctx_config, "fonts", fonts[i], "", NULL);
    if (font_path)
    {
      if (font_path[0])
        ctx_load_font_file (ctx, fonts[i], font_path);
      free (font_path);
    }
  }
#endif
  float old_line_spacing = line_spacing;
  float old_baseline = baseline;

  font_size = cfg_f32_full (&ctx_config, "ctx", "font_size", 19.5f, "font size in pixels");
  line_spacing = cfg_f32_full (&ctx_config, "fonts", "line_spacing", 1.0f, NULL);
  baseline = cfg_f32_full (&ctx_config, "fonts", "baseline", 0.8f, NULL);
  if (line_spacing != old_line_spacing ||
      baseline != old_baseline)
  {
    set_new_font_size (ctx, font_size);
  }
}

static int no_save_config = 0;
extern int ctx_vt_enable_ligatures;

static void load_settings (Ctx *ctx)
{
  load_palette ();
  load_fonts (ctx);

        cfg_boolean_full (&ctx_config, "", "auto_hide_mouse", 1, NULL);
  ctx_vt_enable_ligatures = cfg_boolean_full (&ctx_config, "fonts", "enable_ligatures", -1,
"this causes consequetive runs of the same foreground color to be\n"
"# shaped by harfbuzz; triggering coding font ligatures for things like:\n"
"#  != </> /> == === <- -> => ~~ 0x000 ");


  if (cfg_boolean_full (&ctx_config, "ctx", "internal_clipboard", -1,
"If set to true disables use of system clipboard in SDL backend, thus making\n"
"# other applications not able to snoop on the contents.")>0)
    ctx_internal_clipboard (ctx, 1);

  cfg_f32_full (&ctx_config, "style", "animation_duration", 0.345f, NULL);

  cfg_boolean_full (&ctx_config, "", "auto_hide_mouse", 1, NULL);
  cfg_f32_full (&ctx_config, "style", "overview_gap", 0.5f, "font relative gap between columns");
  cfg_f32_full (&ctx_config, "style", "overview_selected_line_width", 1.0f, NULL);
  cfg_ctx_color_full (&ctx_config, "style", "tile_bg", ctx, "#0027", NULL);
  cfg_ctx_color_full (&ctx_config, "style", "tile_fg", ctx, "#f4", NULL);
  cfg_ctx_color_full (&ctx_config, "style", "overview_bg_shade", ctx, "#0b", NULL);
  cfg_ctx_color_full (&ctx_config, "style", "overview_fg", ctx, "#ff", NULL);
 cfg_ctx_color_full (&ctx_config, "style", "overview_fg_selected", ctx, "#ff", NULL);
 dock_dim = cfg_f32_full (&ctx_config, "style", "dock_width", 4.0f, NULL);
  cfg_f32_full (&ctx_config, "style", "overview_gap", 0.5f, "font relative gap between columns");


  cfg_f32_full (&ctx_config, "fonts", "baseline", 0.78f, NULL);
  cfg_f32_full (&ctx_config, "style", "window_radius", 0.5f, NULL);
  cfg_f32_full (&ctx_config, "style", "border_width", 0.5f, NULL);
  cfg_f32_full (&ctx_config, "style", "resize_border_width", 0.5f, NULL);




 cfg_ctx_color_full (&ctx_config, "style", "window_border_focused", ctx, "#040404", NULL);
 cfg_ctx_color_full (&ctx_config, "style", "window_border", ctx, "#0f", NULL);

 cfg_ctx_color_full (&ctx_config, "style", "window_close", ctx, "#a00", NULL);
 cfg_ctx_color_full (&ctx_config, "style", "window_fg", ctx, "#7f", NULL);
 cfg_ctx_color_full (&ctx_config, "style", "window_fg_active", ctx, "#ff", NULL);

 cfg_f32_full (&ctx_config, "style", "border_shadow_x", 0.0f, NULL);
 cfg_f32_full (&ctx_config, "style", "border_shadow_y", 0.0f, NULL);
 cfg_ctx_color_full (&ctx_config, "style", "window_shadow", ctx, "#00000077", NULL);
 cfg_ctx_color_full (&ctx_config, "style", "wallpaper", ctx, "#051010", NULL);

 cfg_ctx_color_full (&ctx_config, "style", "debug", ctx, "#ff000077", NULL);
 animation_duration = cfg_f32_full (&ctx_config, "style", "animation_duration", 0.345f, NULL);
}



void ctx_client_titlebar_drag (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_n (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_w (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_e (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_s (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_nw (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_ne (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_sw (CtxEvent *event, void *data, void *data2);
void ctx_client_resize_se (CtxEvent *event, void *data, void *data2);
void ctx_client_close (CtxEvent *event, void *data, void *data2);


void client_bigger (CtxEvent *event, void *data, void *data2)
{
  if (is_big)
  {
    is_big = 0;
  }
  else
  {
    is_big = 2;
  }
  ctx_queue_draw (event->ctx);
}


typedef enum {
  DECOR_NONE      = 0,
  DECOR_TITLEBAR  = (1<<0),
  DECOR_FOCUSED   = (1<<1), 
  DECOR_RESIZABLE = (1<<2), 
  DECOR_MOVABLE   = (1<<3),
  DECOR_CLOSE     = (1<<4),
  DECOR_MAXIMIZE  = (1<<5)
} decor_flag;

void decorate_client (Ctx *ctx, CtxClient *client, int client_x, int client_y, int flags)
{
  int is_focused = (flags & DECOR_FOCUSED) != 0;
  int id = ctx_client_id (client);
  //int flags = ctx_client_flags (client);

  float rbw = font_size * cfg_f32 (&ctx_config, "style", "resize_border_width");
  float bw  = font_size * cfg_f32 (&ctx_config, "style", "border_width");
  float sx  = font_size * cfg_f32 (&ctx_config, "style", "border_shadow_x");
  float sy  = font_size * cfg_f32 (&ctx_config, "style", "border_shadow_y");

  //int client_x = ctx_client_x (ctx, id);
  //int client_y = ctx_client_y (ctx, id);
  int client_width = ctx_client_width (ctx, id);
  int client_height = ctx_client_height (ctx, id);
    float window_radius = font_size * cfg_f32 (&ctx_config, "style", "window_radius");



  if (is_focused)
    cfg_ctx_color (&ctx_config, "style", "window_border_focused", ctx);
  else
    cfg_ctx_color (&ctx_config, "style", "window_border", ctx);

  if (flags & DECOR_TITLEBAR)
  {
    ctx_rectangle (ctx, client_x - bw, client_y - font_size,
                      client_width + bw * 2, font_size);

    if (flags & DECOR_MOVABLE)
    {
      ctx_listen (ctx, CTX_DRAG, ctx_client_titlebar_drag, client, NULL);
      ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_ALL);
    }

    if (window_radius > 0.01)
    {
      ctx_reset_path (ctx);
      ctx_arc (ctx, client_x - bw + window_radius, client_y - font_size + window_radius, window_radius, M_PI, M_PI * 1.5f, 0);
      ctx_arc (ctx, client_x + bw + client_width - window_radius, client_y - font_size + window_radius, window_radius, M_PI * 1.5f, M_PI * 2, 0);
      ctx_line_to (ctx, client_x + bw + client_width, client_y);
      ctx_line_to (ctx, client_x - bw, client_y);
    }

    ctx_fill (ctx);

    if (is_focused)
    {
    int pos = 1;
    if (flags & DECOR_CLOSE)
    {
      ctx_rectangle (ctx, client_x + client_width - font_size * pos,
                     client_y - font_size, font_size, font_size);
      ctx_listen (ctx, CTX_CLICK, ctx_client_close, client, NULL);
      ctx_listen_set_cursor (ctx, CTX_CURSOR_CROSSHAIR);
      //ctx_fill (ctx);
      ctx_reset_path (ctx);

      ctx_move_to (ctx, client_x + client_width - font_size * pos,
                   client_y - 0.2 * font_size);
      ctx_save (ctx);



      cfg_ctx_color (&ctx_config, "style", "window_close", ctx);
      ctx_text (ctx, "X");
      ctx_restore (ctx);
      pos ++;
    }

    if (flags & DECOR_MAXIMIZE)
    {
      ctx_rectangle (ctx, client_x,
                     client_y - font_size, font_size, font_size);
      ctx_listen (ctx, CTX_CLICK, client_bigger, client, NULL);
      ctx_listen_set_cursor (ctx, CTX_CURSOR_HAND);
      //ctx_fill (ctx);
      ctx_reset_path (ctx);
      
      ctx_move_to (ctx, client_x, client_y - 0.2 * font_size);
      ctx_save (ctx);
      cfg_ctx_color (&ctx_config, "style", "window_fg", ctx);
      ctx_text (ctx, "O");
      ctx_restore (ctx);
      //pos ++;
    }
    }

  }



  if (bw > 0.0f)
  {
    ctx_rectangle (ctx, client_x - bw, client_y,
                        bw, client_height);
    ctx_fill (ctx);
    ctx_rectangle (ctx, client_x + client_width, client_y,
                        bw, client_height);
    ctx_fill (ctx);
    if (window_radius > 0.01)
    {
      ctx_reset_path (ctx);
      ctx_arc (ctx, client_x - bw + window_radius, client_y + client_height + bw - window_radius,
               window_radius, M_PI, M_PI/2, 1);
      ctx_arc (ctx, client_x + bw + client_width - window_radius,
               client_y + client_height + bw - window_radius,
               window_radius,
               M_PI/2, 0, 1);
      ctx_line_to (ctx, client_x + bw + client_width, client_y + client_height);
      ctx_line_to (ctx, client_x - bw, client_y + client_height);
    }
    else
    {
      ctx_rectangle (ctx, client_x - bw, client_y + client_height,
                          client_width + bw * 2, bw);
    }
    ctx_fill (ctx);
    if ((flags & DECOR_TITLEBAR)==0)
    {

    if (window_radius > 0.01)
    {
      ctx_reset_path (ctx);
      ctx_arc (ctx, client_x - bw + window_radius, client_y - bw + window_radius, window_radius, M_PI, M_PI * 1.5f, 0);
      ctx_arc (ctx, client_x + bw + client_width - window_radius, client_y - bw + window_radius, window_radius, M_PI * 1.5f, M_PI * 2, 0);
      ctx_line_to (ctx, client_x + bw + client_width, client_y);
      ctx_line_to (ctx, client_x - bw, client_y);
    }
    else
    {
      ctx_rectangle (ctx, client_x - bw, client_y - bw,
                        client_width + bw * 2, bw);
    }

      ctx_fill (ctx);
    }
  }

  // shadow - only for active window
  if (fabsf(sx) > 0.0f && is_focused)
  {


    cfg_ctx_color (&ctx_config, "style", "window_shadow", ctx);


    if (sx > 0)
    {
      ctx_rectangle (ctx, client_x + client_width + bw, client_y + sy - font_size,
                           sx, client_height + font_size + bw);
      ctx_fill (ctx);
      ctx_rectangle (ctx, client_x + sx, client_y + client_height + bw,
                           client_width - sx + bw, sy);
      ctx_fill (ctx);
    }
  }

  if (rbw > 0.0f && is_focused && ((flags & DECOR_RESIZABLE) != 0))
  {
        ctx_rgba(ctx, 1,0,0,0.5);
        ctx_rectangle (ctx, client_x - rbw, client_y,
                            rbw, client_height);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_w, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_W);
        ctx_reset_path (ctx);
        ctx_rectangle (ctx, client_x + client_width, client_y,
                            rbw, client_height);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_e, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_E);
        ctx_reset_path (ctx);

        ctx_rectangle (ctx, client_x, client_y + client_height,
                            client_width, rbw);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_s, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_S);
        ctx_reset_path (ctx);

        ctx_rectangle (ctx, client_x - rbw, client_y + client_height,
                            rbw, rbw);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_sw, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_SW);
        ctx_reset_path (ctx);

        ctx_rectangle (ctx, client_x + client_width, client_y + client_height,
                            rbw, rbw);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_se, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_SE);
        ctx_reset_path (ctx);

        ctx_rectangle (ctx, client_x, client_y - font_size - rbw,
                            client_width, rbw);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_n, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_N);
        ctx_reset_path (ctx);

        ctx_rectangle (ctx, client_x - rbw, client_y - font_size - rbw,
                            rbw, rbw);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_nw, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_NW);
        ctx_reset_path (ctx);

        ctx_rectangle (ctx, client_x + client_width, client_y - font_size - rbw,
                            rbw, rbw);
        ctx_listen (ctx, CTX_DRAG, ctx_client_resize_ne, client, NULL);
        ctx_listen_set_cursor (ctx, CTX_CURSOR_RESIZE_NE);
        ctx_reset_path (ctx);
  }



  if (flags & DECOR_TITLEBAR)
  {
        // draw title
        ctx_save (ctx);
        if (is_focused)
          cfg_ctx_color (&ctx_config, "style", "window_fg_active", ctx);
        else
          cfg_ctx_color (&ctx_config, "style", "window_fg", ctx);

        ctx_text_align (ctx, CTX_TEXT_ALIGN_CENTER);
        ctx_move_to (ctx, client_x + client_width/2, client_y - font_size * 0.2f);
        ctx_text (ctx, ctx_client_get_title (ctx, id));
        ctx_restore (ctx);
  }



}

void window_manager (Ctx *ctx, float delta_s)
{
  cfg_ctx_color (&ctx_config, "style", "wallpaper", ctx);
  ctx_paint(ctx);

  CtxList *clients = ctx_clients (ctx);
  //int n_clients = ctx_list_length (clients);

  int active_id = ctx_clients_active (ctx);
  CtxClient *active = active_id>=0?ctx_client_by_id (ctx, active_id):NULL;

  for (CtxList *l = clients; l; l = l->next)
  {
    CtxClient *client = l->data;
    int id = ctx_client_id (client);
    int flags = ctx_client_flags (client);
    if (flag_is_set(flags, CTX_CLIENT_MAXIMIZED))
    {
      if (client == active)
        ctx_client_draw (ctx, client, ctx_client_x(ctx, id), ctx_client_y(ctx, id));
      else
        ctx_client_use_images (ctx, client);

    } else
    {
        int client_x = ctx_client_x (ctx, id);
        int client_y = ctx_client_y (ctx, id);

        ctx_client_draw (ctx, client, client_x, client_y);
        decorate_client (ctx, client, client_x, client_y,
          DECOR_TITLEBAR |
          DECOR_CLOSE |
          DECOR_MAXIMIZE |
          DECOR_RESIZABLE | 
          DECOR_MOVABLE | 
          ((active == client) * DECOR_FOCUSED) );

    }

  }

  draw_panel (ctx);
}

static void raise_to_second (CtxEvent *event, void *data1, void *data2)
{
  CtxClient *client = data1;
  Ctx *ctx = event->ctx;
  ctx_client_raise_almost_top (ctx, ctx_client_id (client));
  ctx_queue_draw (ctx);
  ctx_client_focus (ctx, ctx_client_id (client));
  event->stop_propagate = 1;
}


typedef struct 
{
  CtxClient *client;
  int   id;
  float x;
  float y;
  float scale;
  float opacity;
  float start_x;
  float start_y;
  float start_scale;
  float start_opacity;
  float target_x;
  float target_y;
  float target_scale;
  float target_opacity;
  float deadline;
  float elapsed;
  uint64_t last_alive;
} AnimatedPos;

#define MAX_CLIENTS 32

AnimatedPos anim_state[MAX_CLIENTS]={{0,}};

static AnimatedPos *get_state(Ctx *ctx, CtxClient *client, uint64_t time)
{
  int id = ctx_client_id (client);
  int oldest = -1;
  uint64_t oldest_time = (1L<<63);
  for (int i = 0; i < MAX_CLIENTS; i++)
  {
    if (anim_state[i].id == id)
    {
      anim_state[i].last_alive = time;
      return &anim_state[i];
    }
    if (anim_state[i].last_alive < oldest_time)
    {
      oldest_time = anim_state[i].last_alive;
      oldest = i;
    }
  }
  for (int i = 0; i < MAX_CLIENTS; i++)
  {
    if (anim_state[i].client == NULL || (i == oldest))
    {
      anim_state[i].client = client;
      anim_state[i].id = id;
      anim_state[i].x = ctx_width (ctx)*0.25;
      anim_state[i].y = ctx_height (ctx)*0.66;
      anim_state[i].last_alive = time;
      return &anim_state[i];
    }
  }
  return &anim_state[MAX_CLIENTS-1];
}

static float easeInOutCubic(float x){
return x < 0.5f ? 4 * x * x * x : 1 - powf(-2 * x + 2, 3.0f) / 2.0;
}

float border_width = 0.1f;

static void four_by_three_layout (Ctx    *ctx,
                                  int     focused,
                                  int     no,
                                  int     client_count,
                                  float   small_part,
                                  float   small_scale,
                                  float  *ret_x,
                                  float  *ret_y,
                                  float  *ret_scale,
                                  float  *ret_opacity)
{
  int height = ctx_height (ctx) * (1.0f-small_part);
  int width  = ctx_width (ctx)/2 - font_size;

  float big_y;
  float small_y;

  big_y = ctx_height (ctx) - (height) - font_size * 0.5f;
  small_y = font_size / 2;

  float single_scale = 1.5;
  float single_x = font_size * border_width;
  float single_y = font_size * single_scale;

  *ret_opacity = 0.0f;

  if (client_count == 1)
  {
    *ret_scale = single_scale;
    *ret_x = single_x;
    *ret_y = single_y;
    *ret_opacity = 1.0f;
    return;
  }
    if (is_big)
    {

      if (focused)
      {
        *ret_scale = single_scale;
        *ret_x = single_x;
        *ret_y = single_y;
        if (is_big == 1)
        {
          *ret_scale = single_scale * ((height + font_size) / height);
          *ret_x = (ctx_width (ctx) -  width * *ret_scale)/2;
          *ret_y = single_y - font_size **ret_scale;
        }
        *ret_opacity = 1.0f;
        return;
      }

      no = client_count - no - 2;
      if (no<0) no = 0;
      *ret_x = ctx_width(ctx) - width * small_scale - font_size/2;
      *ret_y = font_size/2 + (height * small_scale + font_size) * no;
      *ret_scale = small_scale;
      *ret_opacity = 1.0f;

      if (is_big == 1)
        *ret_opacity = 0.0f;

      return;
    }

    if (no == client_count-1)
    {
      *ret_x = font_size * 0.5;
      *ret_y = big_y;
      *ret_opacity = 1.0f;
    }
    else if (no == client_count-2)
    {
      *ret_x = ctx_width(ctx) - width - font_size/2;
      *ret_y = big_y;
      *ret_opacity = 1.0f;
    }
    else if (client_count < 16)
    {

    #if 1
    if (no == client_count-3)
    {
      *ret_x = (ctx_width(ctx) / 2) * (0) * small_scale + font_size * 0.5;
      *ret_y = small_y;
      *ret_scale = small_scale;
    }
    else if (client_count<13 && no == client_count-4)
    {
      *ret_x = (ctx_width(ctx) / 2) * (1) * small_scale + font_size * 0.5;
      *ret_y = small_y;
      *ret_scale = small_scale;
    }
    else if (client_count<10 && no == client_count-5)
    {
      *ret_x = (ctx_width(ctx) / 2) * (2) * small_scale + font_size * 0.5;
      *ret_y = small_y;
      *ret_scale = small_scale;
    }
    else if (client_count<7 && no == client_count-6)
    {
      *ret_x = (ctx_width(ctx) / 2) * (3) * small_scale + font_size * 0.5;
      *ret_y = small_y;
      *ret_scale = small_scale;
    }
    else
    {
      int base = 3;
      int row, col;
      if (client_count >=13)
      {
        base = 1;
        row = (client_count-4-no) % 2;
        col = (client_count-4-no) / 2;
      }
      else if (client_count >=10)
      {
        base = 2;
        row = (client_count-5-no) % 2;
        col = (client_count-5-no) / 2;
      }
        else
      {
        base = 3;
        row = (client_count-6-no) % 2;
        col = (client_count-6-no) / 2;
      }
      *ret_x = (ctx_width(ctx) / 2) * (base) * small_scale + font_size * 0.5 + (width*small_scale/2) * col;
      *ret_y = small_y + height * (small_scale/2) * row;
      *ret_scale = small_scale/2;
    }

      *ret_opacity = 1.0f;
    }
    else 
    {
      float rows = 2;
      if (client_count < 7)
        rows = 1;
      else if (client_count < 9)
        rows = 1.5;
      else if (client_count < 19)
        rows = 2;
      else if (client_count < 30)
        rows = 3;
      
      small_scale /= rows;
      int cols = ctx_width (ctx) / (width * small_scale);
      int irows = ((client_count-2 + (cols-1)) / cols);
      int i = (client_count-2-1-no);

      int col = i % cols;
      int row = i / cols;

      //col = cols - col - 1;
      row = irows - row - 1;

      *ret_x = (ctx_width(ctx) / 2) * (col) * small_scale + font_size * 0.5;
      *ret_y = small_y + (height * small_scale + font_size) * row ;
      *ret_scale = small_scale;
      *ret_opacity = 1.0f;
    }
    #endif



}

void four_by_three (Ctx *ctx, float delta_s)
{
  border_width  = cfg_f32 (&ctx_config, "style", "border_width");

  cfg_ctx_color (&ctx_config, "style", "wallpaper", ctx);


  ctx_paint(ctx);

  CtxList *clients = ctx_clients (ctx);
  int client_count = ctx_list_length (clients);

  int active_id = ctx_clients_active (ctx);
  CtxClient *active = active_id>=0?ctx_client_by_id (ctx, active_id):NULL;
  int no = 0;

#if 0
  float small_part  = (1.0-0.618033f);
  float small_scale = 0.55;
#else
  float small_scale = 0.485f;
  float small_part  = 0.35f;
#endif

  int height = ctx_height (ctx) * (1.0f-small_part);
  int width  = ctx_width (ctx)/2 - font_size;

  uint64_t ticks = ctx_ticks ();

  for (CtxList *l = clients; l; l = l->next)
  {
    CtxClient *client = l->data;

    if (l->next && l->next->next == NULL)
      {
        second = l->data;
        client = l->next->data;
      }
    else if (l->next == NULL && second)
      {
        client = second;
      }


    int id = ctx_client_id (client);
    //int flags = ctx_client_flags (client);
    AnimatedPos *astate = get_state(ctx, client, ticks);

    //int client_x = ctx_client_x (ctx, id);
    //int client_y = ctx_client_y (ctx, id);
    float client_x;
    float client_y;

    if (ctx_client_width (ctx, id) != width ||
        ctx_client_height (ctx, id) != height)
    {
      ctx_client_unmaximize (ctx, id);
      ctx_client_resize (ctx, id, width, height);
    }

    float scale = 1.0f;
    float opacity = 1.0f;

    four_by_three_layout (ctx, active == client, no, client_count, small_part, small_scale, &client_x, &client_y, &scale, &opacity);

    if (astate->target_x != client_x ||
        astate->target_y != client_y ||
        astate->target_scale != scale ||
        astate->target_opacity != opacity)
    {
      astate->start_x = astate->x;
      astate->start_y = astate->y;
      astate->start_scale = astate->scale;
      astate->start_opacity = astate->opacity;
      astate->target_x = client_x;
      astate->target_y = client_y;
      astate->target_scale = scale;
      astate->target_opacity = opacity;
      astate->deadline = animation_duration;
      astate->elapsed = - delta_s; // to avoid much time having passed when just triggerd
    }
    //fprintf (stderr, "{%f}", delta_s);

    if (astate->elapsed > astate->deadline)
    {
      astate->scale = astate->start_scale = astate->target_scale;
      astate->x = astate->start_x = astate->target_x;
      astate->y = astate->start_y = astate->target_y;
      astate->opacity = astate->start_opacity = astate->target_opacity;
    }
    else
    {
      float t = easeInOutCubic (astate->elapsed / astate->deadline);
      //float t = (astate->elapsed / astate->deadline);
      if (t < 0)
        t = 0.0f;
      astate->x     = astate->start_x     * (1.0f - t) + t * astate->target_x;
      astate->y     = astate->start_y     * (1.0f - t) + t * astate->target_y;
      astate->scale = astate->start_scale * (1.0f - t) + t * astate->target_scale;
      astate->opacity = astate->start_opacity * (1.0f - t) + t * astate->target_opacity;
      ctx_queue_draw (ctx);
    }
    astate->elapsed += delta_s;
    
    client_x = astate->x;
    client_y = astate->y;
    scale = astate->scale;

    if (astate->opacity > 0.0f)//is_big == 0 ||
        //(is_big == 2 &&  ((active == client) || no < 4   )) ||
        //(is_big == 1 && (active == client)))
    {
      ctx_save (ctx);
      if (astate->opacity != 1.0f)
      {
        ctx_global_alpha (ctx, astate->opacity);
      }
      ctx_translate (ctx, client_x, client_y);
      ctx_scale (ctx, scale, scale);
  
      ctx_client_draw (ctx, client, 0, 0);
      decorate_client (ctx, client, 0, 0, 
         DECOR_TITLEBAR |
         DECOR_CLOSE |
         //DECOR_MOVABLE |  // causes focus to work on titlebars
         DECOR_MAXIMIZE |
         (active == client) * DECOR_FOCUSED);
  
      if (scale < 1.0f)
      {
        ctx_rectangle (ctx, 0, 0, width, height);
        ctx_listen (ctx, CTX_DRAG, raise_to_second, client, NULL);
        ctx_reset_path (ctx);
      }
  
      ctx_restore (ctx);
    }
    else
    {
      ctx_client_use_images (ctx, client);
    }

    no++;
  }

  if (client_count > 1 && !is_big)
  {

    ctx_rectangle (ctx, ctx_width (ctx)/2 - font_size * 1.5, 
    ctx_height(ctx)-font_size*3
    , font_size * 3, font_size * 3);
    ctx_listen (ctx, CTX_CLICK, raise_to_top, second, NULL);
    ctx_reset_path (ctx);
    ctx_rgba (ctx, 1,1,1,1);
    icon (ctx, ctx_width(ctx)/2 - font_size * 1,
               ctx_height(ctx)-font_size * 2,
        2 * font_size, 2 * font_size, "switch", 

"lineWidth 0.2 lineCap round lineJoin round rgba 0 0 0 0.25 M 0.2 0.45 m 0.15 0.15 l -0.15 -0.15 l 0.15 -0.15 l -0.15 0.15 l 0.6 0.0 l -0.15 0.15 l 0.15 -0.15 l -0.15 -0.15 l 0.15 0.15 stroke rgba 1 1 1 0.5 lineWidth 0.08 M 0.2 0.45 m 0.15 0.15 l -0.15 -0.15 l 0.15 -0.15 l -0.15 0.15 l 0.6 0.0 l -0.15 0.15 l 0.15 -0.15 l -0.15 -0.15 l 0.15 0.15 stroke ");


  }

  draw_panel (ctx);

  if (client_count == 1)//< 6)
  {
    help_elapsed += delta_s;
    if (help_elapsed < 30.0f)
    {
    float x0 = width * 1.5f + font_size * border_width * 2 + font_size * 0.2;

    if (ui_button_coords (ui, x0, ctx_height (ctx) - font_size * 23*2,
                              font_size * 6, font_size * 3, "dismiss", 1, &help_elapsed))
    {
       help_elapsed += 40;
    }
             

    ctx_save (ctx);
    ctx_rgba (ctx, 1,1,1,0.5);
    ctx_move_to (ctx, x0, ctx_height (ctx) - font_size * 20*2);
    ctx_wrap_left (ctx, x0);
    ctx_wrap_right (ctx, ctx_width (ctx) - font_size * 0.5);
    ctx_text (ctx,
"Keyboard actions:\n\n"
"alt 1..9 switch to window\n\n"
"shift-ctrl-T new window\n\n"
"shift-ctrl-W close window\n\n"
"shift-ctrl-Q quit\n\n"
"shift-ctrl-M (un)maximize\n\n"
"shift-ctrl-F (un)fullscreen\n\n"
"shift-ctrl-J swap front windows\n\n"
"shift-ctrl-C copy selection\n\n"
"shift-ctrl-V paste\n\n"
"shift-ctrl+= increase font size\n\n"
"shift-ctrl-- decrease font size\n\n"
"shift-ctrl-up scroll up\n\n"
"shift-ctrl-down scroll down\n\n"
"\n"
"configuration is stored in ~/.config/ctx/ctx.ini\n"
);
    ctx_restore (ctx);
    }
  }

}

void tabs_and_overview (Ctx *ctx, float delta_s);


#if CTX_BIN_BUNDLE
int ctx_terminal_main (int argc, char **argv)
#else
int main (int argc, char **argv)
#endif
{
  execute_self = argv[0];
  const char *commandline = NULL;
  float global_scale = 1.0;

  {
    ctx_config_timestamp = path_modtime (ctx_term_config_path());
    cfg_load (&ctx_config, ctx_term_config_path (),
"# ctx-configuration, this file is kept in sync with ctx at runtime\n"
);
    {
      char *ver = cfg_string (&ctx_config, "ctx", "version");
      if (ver)
      {
        if (!strcmp (ver, CTX_VERSION_STRING))
        {
           // mostly remove old things?
           // we for that no specific check is needed just remove any old cruft?
        }
        free (ver);
      }
    }

    cfg_add_category (&ctx_config, "ctx", "ctx settings");
    cfg_set_string (&ctx_config, "ctx", "version", CTX_VERSION_STRING, NULL);
    cfg_add_category (&ctx_config, "fonts",
"Loaded \"system\" fonts, fallbacks exist for Regular and Mono when left as empty\n# strings. Valid values are full paths to OTF, TTF or ctx font files.");
    cfg_add_category (&ctx_config, "style", "Visual/UI appearance.");
    cfg_add_category (&ctx_config, "palette", "terminal color palette");
    //cfg_add_category (&ctx_config, "system", "misc global settings");
  }

  int experimental = cfg_boolean_full (&ctx_config, "", "experimental", -1, "enable unfinished ui parts");
  if (experimental)
  {
    volume = cfg_f32 (&ctx_config, "system", "volume");
    backlight = cfg_f32_full (&ctx_config, "system", "backlight", 100.0f, "software or hardware backlight");
  }
  
  layout_no = cfg_int_full (&ctx_config, "", "layout_no", 2, "different ui modes");

  int height = cfg_int_full (&ctx_config, "ctx", "height", 1080, NULL);
  int width  = cfg_int_full (&ctx_config, "ctx", "width",  1920, "window width in pixels, -1 for backend default");
  int fullscreen = cfg_boolean_full (&ctx_config, "ctx", "fullscreen", -1, NULL);
  int monitor_config = cfg_boolean_full (&ctx_config, "", "monitor_config", 1, "at a performance cost, check for timestamp changes on config file");
  int cols = -1;


  if (getpid () == 1)
  {
    int ign;
    ign = system ("pkill plymouth"); // might be hogging stdin
    ign = system ("mount -o remount,rw /");
    ign = system ("mount -a");
    if (ign) {};
  }

  for (int i = 1; argv[i]; i++)
  {
    if (!strcmp (argv[i], "--help"))
    {
      printf ("ctx-terminal [--width px] [--height px] [--cols target columns] [--font-size size] [-e [--] command]\n");
      return 0;
    }
    else if (!strcmp (argv[i], "-e"))
    {
      commandline = argv[i+1];
      i++;
      if (!strcmp (commandline, "--"))
      {
        i++;
        commandline_argv_start = i;
      }
      terminal_no_new_tab = 1;
    }
    else if (!strcmp (argv[i], "--fullscreen"))
      fullscreen = 1;
    else if (!strcmp (argv[i], "--no-save-config"))
      no_save_config = 1;
    else if (!strcmp (argv[i], "--width"))
        width = consume_float (argv, &i);
    else if (!strcmp (argv[i], "--height"))
        height = consume_float (argv, &i);
    else if (!strcmp (argv[i], "--cols"))
        cols = consume_float (argv, &i);
    else if (!strcmp (argv[i], "--font-size"))
        font_size = consume_float (argv, &i) * global_scale;
    else if (argv[i][0]=='-')
    {
      switch (argv[i][1])
      {
        case 's': font_size = consume_float (argv, &i) * global_scale; break;
        case 'c': cols = consume_float (argv, &i); break;
        case 'w': width = consume_float (argv, &i); break;
        case 'h': height = consume_float (argv, &i); break;
      }
    }
  }

  if (font_size > 0 && width <0 && cols < 0)
  {
    cols = 80;
    width = font_size * cols / 2;
    height = font_size * 25;
  }

  ctx  = ctx_new (width, height, NULL);
  width = ctx_width (ctx);
  height = ctx_height (ctx);

    if (layout_no == 1)
      ctx_set_focus_cb (ctx, wm_focus_cb, NULL);
    else
      ctx_set_focus_cb (ctx, NULL, NULL);
  
  if (fullscreen)
    ctx_set_fullscreen (ctx, 1);

  load_settings(ctx);


#if CTX_TERM
  if (ctx_backend_type (ctx) == CTX_BACKEND_TERM && font_size <= 0)
  {
    font_size = ctx_term_get_cell_height (ctx);
  }
#endif

  if (cols <= 0)
  {
    //if (((double)(width))/height < (16.0/9.0))
    //  cols = 80;
    //else
      cols = 80;
  }

  if (font_size < 0)
    font_size = floorf (width / cols  * 2 / 1.5);

  //void *itk = NULL;

  ctx_font_size (ctx, font_size);

  if (commandline_argv_start)
  {
    ctx_client_maximize (ctx, add_tab_argv (ctx, &argv[commandline_argv_start], 1));
  }
  else
  {
    if (!commandline)
      commandline = ctx_find_shell_command();
    ctx_client_maximize (ctx, add_tab (ctx, commandline, 1));
  }

#if GNU_C
  int mt = ctx_add_timeout (ctx, 1000 * 20, malloc_trim_cb, NULL);
#endif


  float prev_ms = ctx_ms (ctx);
  int cfg_has_reloaded = 1;

  ui = ui_new2 (ctx);

  while (ctx_clients (ctx) && !ctx_has_exited (ctx))
    {
      //int n_clients = ctx_list_length (ctx_clients (ctx));
      float ms = ctx_ms (ctx);
      float delta_s = (ms - prev_ms)/1000.0f;
      prev_ms = ms;

      full_time += delta_s;
      if (cfg_has_reloaded)
      {
        load_settings (ctx);
        cfg_has_reloaded = 0;
      }

      {
        static int lock_control = 0;
        lock_control--;

        if (lock_control < 0)
        {
          int lastval = locked;
          if (access("/tmp/ctx.lock", R_OK) == F_OK)
          {
            locked = 1;
          }
          else
          {
            locked = 0;
          }
          if (locked != lastval)
          {
            ctx_queue_draw (ctx);
            if (locked)
            {
              add_tab (ctx, CTX_TERM_UNLOCK_SH, 0);
            }
          }
          lock_control = 50;
        }

      }

      if (ctx_need_redraw(ctx))
      {
        float elapsed = ctx_start_frame (ctx);
        ui_start_frame (ui);

        if (!no_save_config && ctx_cfg_dirty != 0.0f)
        {
          ctx_cfg_dirty += elapsed;
          if (ctx_cfg_dirty > 4.0)
          {
            cfg_save (&ctx_config, ctx_term_config_path ());
            ctx_config_timestamp = path_modtime (ctx_term_config_path());
          }
        }

#if 1
        if (monitor_config &&
            (ctx_config_timestamp != path_modtime (ctx_term_config_path())))
        {
          ctx_wait_for_renderer (ctx);
          usleep (200 * 1000);
          ctx_config_timestamp = path_modtime (ctx_term_config_path());
          cfg_load (&ctx_config, ctx_term_config_path (), NULL);
          monitor_config = cfg_boolean (&ctx_config, "", "monitor_config");
          font_size = cfg_f32 (&ctx_config, "ctx", "font_size");
          line_spacing = cfg_f32 (&ctx_config, "ctx", "line_spacing");
          baseline = cfg_f32 (&ctx_config, "ctx", "baseline");
          dock_dim = cfg_f32 (&ctx_config, "style", "dock_width");
          animation_duration = cfg_f32 (&ctx_config, "style", "animation_duration");

          #if 0
          experimental = cfg_boolean (&ctx_config, "", "experimental");
          if (experimental)
          {
            backlight = cfg_f32 (&ctx_config, "system", "backlight");
            volume = cfg_f32 (&ctx_config, "system", "volume");
          }
          for (CtxList *l = ctx_clients (ctx); l; l = l->next)
          {
            CtxClient *client = l->data;
            VT *vt = ctx_client_vt (client);
            vt_set_font_size (vt, font_size);
            vt_set_baseline (vt, baseline);
            vt_set_line_spacing (vt, line_spacing);
            vt_recompute_cols (vt);
          }
          ctx_queue_draw (ctx);
          #endif
          cfg_has_reloaded = 1;
        }
  #endif

        //ctx_rgb (ctx, 0.086,0.086,0.113);
        ctx_font_size (ctx, font_size);
        ctx_save (ctx);
#if ENABLE_ROTATE
        ctx_translate (ctx, ctx_width(ctx)/2, ctx_height(ctx)/2);
        ctx_rotate (ctx, global_rotation);
        ctx_translate (ctx, -ctx_width(ctx)/2, -ctx_height(ctx)/2);
#endif
        //ctx_rectangle (ctx, 0, 0, ctx_width (ctx), ctx_height (ctx));
        //ctx_fill (ctx);
        //ctx_paint (ctx);

        if (locked)
        {
          ctx_term_lock_screen (ctx);
        }
        else
        {
          switch (layout_no)
          {
            case 0:
            default:
              tabs_and_overview (ctx, elapsed);
              break;
            case 1:
              window_manager (ctx, elapsed);
              break;
            case 2:
              four_by_three (ctx, elapsed);
              break;
          }
        }

        if (!in_overview || locked)
          ctx_osk_draw (ctx);
        ctx_restore (ctx);

        if (backlight != 100.0f)
        {
          ctx_rgba (ctx, 0.0,0.0,0.0, (100.0f-backlight)/100.0f);
          ctx_paint (ctx);
        }

        ctx_listen (ctx, CTX_KEY_PRESS, terminal_key_any, NULL, NULL);
        ctx_listen (ctx, CTX_KEY_DOWN,  terminal_key_any, NULL, NULL);
        ctx_listen (ctx, CTX_KEY_UP,    terminal_key_any, NULL, NULL);

        ctx_rectangle (ctx, 0, 0, ctx_width (ctx), ctx_height (ctx));


        if (cfg_boolean (&ctx_config, "", "auto_hide_mouse"))
          ctx_listen (ctx, CTX_MOTION, terminal_show_cursor, NULL, NULL);
        if (!in_overview || locked)
          ctx_listen (ctx, CTX_TAP_AND_HOLD, terminal_long_tap, NULL, NULL);
        ctx_reset_path (ctx);

        ui_end_frame (ui);
        ctx_end_frame (ctx);
      }
      else
      {
        ctx_handle_events (ctx);
      }

     {
       int active_id = ctx_clients_active (ctx);
       CtxClient *active = active_id>=0?ctx_client_by_id (ctx, active_id):NULL;
       if (active)
         terminal_update_title (ctx_client_title (active));
     }

    }

  ui_destroy (ui);

#if GNU_C
  ctx_remove_idle (ctx, mt);
#endif

  if (!no_save_config)
  {
    cfg_save (&ctx_config, ctx_term_config_path ());
  }

  ctx_destroy (ctx);

  if (getpid () == 1)
  {
    if (system ("reboot --force")){};
  }

  return 0;
}


float leave_overview = 0.0f;
float overview_t     = 0.0f;

static void overview_init (Ctx *ctx);

static void overview_event (CtxEvent *event, void *a, void *b)
{
     in_overview = !in_overview;
     //overview_init (event->ctx);
     if (in_overview)
     {
       overview_t = 0.0f;
     }
     else
     {
       leave_overview = animation_duration;
       full_time = 0.0f;
     }
     ctx_queue_draw (event->ctx);
     ctx_event_stop_propagate (event);
}

static void overview_corner_event (CtxEvent *event, void *a, void *b)
{
  int width = ctx_width (event->ctx);
  static int prev_corner = 0;

  int in_corner = (((int)event->x) == width -1) && (((int)event->y) == 0);

  if (in_corner && !prev_corner)
  {
    overview_event (event, a, b);
  }
  prev_corner = in_corner;
}

typedef struct {
  float *ptr;
  int    state;
  float  min;
  float  max;
  float  value;
  float  widget_x;
  float  widget_width;
} FloatSlider;

int sliders_changed = 0;

static void volume_cb (CtxEvent *event, void *a, void *b)
{
  FloatSlider *slider = a;
  int is_font_size = slider->ptr == &font_size;
  switch (event->type)
  {
     case CTX_DRAG_PRESS:
       slider->state += 1;
       if (is_font_size)
         force_full_size = 1;
       if (slider->state > 1)
         slider->state = 0;
       slider->value = *slider->ptr;
       ctx_queue_draw (event->ctx);
       break;
     case CTX_DRAG_MOTION:
       slider->value -= event->delta_y / (ctx_height(ctx)*0.25) * (slider->max-slider->min);
       *slider->ptr = slider->value;
       if (*slider->ptr < slider->min)
         *slider->ptr = slider->min;
       if (*slider->ptr > slider->max)
         *slider->ptr = slider->max;

       if (is_font_size)
       {
         cfg_set_f32 (&ctx_config, "ctx", "font_size", font_size, NULL);
         *slider->ptr = (slider->value);
         for (CtxList *l = ctx_clients (ctx); l; l = l->next)
         {
           CtxClient *client = l->data;
           VT *vt = ctx_client_vt (client);
           vt_set_font_size (vt, font_size);
           vt_recompute_cols (vt);
         }
       }
       ctx_queue_draw (event->ctx);
       sliders_changed = 1;
       break;
     case CTX_DRAG_RELEASE:
       //slider->state = 0;
       if (is_font_size)
         force_full_size = 0;
       ctx_queue_draw (event->ctx);
       break;
     default:
       break;
  }
  ctx_event_stop_propagate (event);
}

static void volume_cb2 (CtxEvent *event, void *a, void *b)
{
  FloatSlider *slider = a;
  switch (event->type)
  {
     case CTX_DRAG_PRESS:
     case CTX_DRAG_MOTION:
       *slider->ptr = (event->x - slider->widget_x) / (slider->widget_width) * (slider->max - slider->min) + slider->min;
       if (*slider->ptr < slider->min)
         *slider->ptr = slider->min;
       if (*slider->ptr > slider->max)
         *slider->ptr = slider->max;

       ctx_queue_draw (event->ctx);
       sliders_changed = 1;
       break;
     case CTX_DRAG_RELEASE:
       ctx_queue_draw (event->ctx);
       break;
     default:
       break;
  }
  ctx_event_stop_propagate (event);
}


static void overview_select_client (CtxEvent *event, void *client, void *data2)
{
  int i;
  int old_active = 0;
  int new_active = 0;

  for (i = 0; i < opos_count; i++)
  {
    if (ctx_client_is_active_tab (event->ctx, opos[i].client))
      old_active = i;
    if (opos[i].client == client)
      new_active = i;
  }
  if (new_active != old_active)
  {
    opos[new_active].x0 = opos[old_active].x0;
    opos[new_active].y0 = opos[old_active].y0;
    opos[new_active].scale0 = 1.0f;
    opos[new_active].w0 = opos[old_active].w0;
    opos[new_active].h0 = opos[old_active].h0;

    opos[old_active].x0 = opos[old_active].x1;
    opos[old_active].y0 = opos[old_active].y1;
    opos[old_active].scale0 = opos[old_active].scale1;


    OverviewPos temp = opos[new_active];
    opos[new_active] = opos[old_active];
    opos[old_active] = temp;

  }
  else
  {
    in_overview = 0;
    leave_overview = animation_duration;
  }


  event->stop_propagate =1;
  ctx_client_focus (event->ctx, ctx_client_id (client));
  //overview_init (ctx);
  ctx_queue_draw (event->ctx);
}

void ctx_dock (Ctx *ctx)
{
  float em = font_size;

  ctx_save (ctx);
  ctx_font_size (ctx, em * 0.9);
  float w = ctx_width (ctx);

  float y = 0;
  float tile_dim = em * dock_dim;
  float x = w - tile_dim;

  ctx_rectangle (ctx, x, y, tile_dim, tile_dim);
  ctx_listen (ctx, CTX_MOTION, overview_corner_event, NULL, NULL);
  ctx_listen (ctx, CTX_PRESS, overview_event, NULL, NULL);

  cfg_ctx_color (&ctx_config, "style", "tile_bg", ctx);
  if (in_overview)
  {
    ctx_fill (ctx);
    ctx_save (ctx);
    cfg_ctx_color (&ctx_config, "style", "tile_fg", ctx);


    icon (ctx, x, y, tile_dim, tile_dim, "overview", "arc 0.5 0.5 0.3 0.0 6.4 0 arc 0.5 0.5 0.2 6.4 0.0 1 fill");
    ctx_fill (ctx);
    ctx_restore (ctx);
  }
  else if (full_time < 4.0f)
  {
    ctx_reset_path (ctx);
    ctx_save (ctx);
    cfg_ctx_color (&ctx_config, "style", "tile_fg", ctx);


    ctx_global_alpha (ctx, 0.5f);
    icon (ctx, x, y, tile_dim, tile_dim, "overview", NULL);
    ctx_fill (ctx);
    ctx_restore (ctx);
  }


  if (!in_overview)
  {
    ctx_restore (ctx);
    return;
  }

  y += tile_dim;

  ctx_rectangle (ctx, x, y, tile_dim, tile_dim);
  ctx_listen (ctx, CTX_PRESS, add_tab_cb, NULL, NULL);
  ctx_fill (ctx);
  ctx_save (ctx);
  cfg_ctx_color (&ctx_config, "style", "tile_fg", ctx);
  icon (ctx, x, y, tile_dim, tile_dim, "terminal", "roundRectangle 0.1 0.1 0.8 0.8 0.1 lineWidth 0.04 stroke M 0.14 0.5 fontSize 0.4 text '$' rectangle 0.36 0.22 0.16 0.33 fill");
  ctx_fill (ctx);
  ctx_restore (ctx);


#if ENABLE_ROTATE
  y += tile_dim;
  ctx_rectangle (ctx, x, y, tile_dim, tile_dim);
  ctx_listen (ctx, CTX_PRESS, rotate_cb, NULL, NULL);
  ctx_fill (ctx);
  ctx_save (ctx);
  cfg_ctx_color (&ctx_config, "style", "tile_fg", ctx);
  icon (ctx, x, y, tile_dim, tile_dim, "rotate", "arc 0.5 0.5 0.2 1.14 6.3 0 L 0.7 0.4 lineWidth 0.1 stroke");
  ctx_fill (ctx);
  ctx_restore (ctx);
#endif

#if 1
  y += tile_dim;
  ctx_rectangle (ctx, x, y, tile_dim, tile_dim);
  ctx_listen (ctx, CTX_PRESS, lock_cb, NULL, NULL);
  ctx_fill (ctx);
  ctx_save (ctx);
  cfg_ctx_color (&ctx_config, "style", "tile_fg", ctx);
  icon (ctx, x, y, tile_dim, tile_dim, "padlock", "M 0.3 0.4 arc 0.5 0.3 0.2 3.14 6.3 0 L 0.7 0.4 lineWidth 0.1 stroke rectangle 0.2 0.4 0.6 0.4 fill");
  ctx_fill (ctx);
  ctx_restore (ctx);
#endif


#define TILE_SLIDER(label, mini, maxi, pntr) \
  { \
    y += tile_dim;\
    float slider_width = tile_dim * 4;\
    ctx_rectangle (ctx, x, y, tile_dim, tile_dim);\
    static FloatSlider cf = {\
      pntr,\
      0,\
      mini,\
      maxi,\
    };\
    cf.widget_x = x - slider_width;\
    cf.widget_width = slider_width;\
    ctx_listen (ctx, CTX_DRAG, volume_cb, &cf, NULL);\
    if (cf.state && cf.ptr != &font_size)\
    {\
      ctx_fill (ctx);\
\
      ctx_save (ctx);\
      ctx_rgba (ctx, 0,1.0,0.3, 1.0f);\
      ctx_rectangle (ctx, x - slider_width , y, slider_width, tile_dim);\
      ctx_listen (ctx, CTX_DRAG, volume_cb2, &cf, NULL);\
      ctx_stroke (ctx);\
      ctx_rectangle (ctx, x - slider_width  - 1 + (*cf.ptr - cf.min) / (cf.max-cf.min) * slider_width, y, 2, tile_dim);\
      ctx_fill (ctx);\
      ctx_restore (ctx);\
    }\
    else\
    {\
      ctx_fill (ctx);\
    }\
    ctx_save (ctx);\
    cfg_ctx_color (&ctx_config, "style", "tile_fg", ctx); \
    char buf[50];\
    sprintf(buf, "%.2f", *cf.ptr);\
    ctx_text_align (ctx, CTX_TEXT_ALIGN_CENTER);\
    ctx_move_to (ctx, x + tile_dim/2, y + tile_dim * 0.33);\
    ctx_text (ctx, label);\
    ctx_move_to (ctx, x + tile_dim/2, y + tile_dim * 0.66);\
    ctx_text (ctx, buf);\
    ctx_restore (ctx);\
  }

  TILE_SLIDER("font size", 8.0f, 42.0f, &font_size);

  if (experimental)
  {

    TILE_SLIDER("volume", 0.0f, 11.0f, &volume);
    TILE_SLIDER("backlight", 5.0f, 100.0f, &backlight);

    if (sliders_changed)
    {
      cfg_set_f32 (&ctx_config, "system", "volume", volume, NULL);
      cfg_set_f32 (&ctx_config, "system", "backlight", backlight, NULL);
      sliders_changed = 0;
    }
  }

  ctx_restore (ctx);
}

static void overview (Ctx *ctx, float anim_t)
{
  CtxList *clients = ctx_clients (ctx);
  float em = ctx_get_font_size (ctx);
  int n_clients = ctx_list_length (clients);


  float gap = cfg_cached_f32 (&ctx_config, "style", "overview_gap") * em;
  float screen_width = ctx_width (ctx) - cfg_cached_f32 (&ctx_config, "style", "dock_width") * em - gap;
  float screen_height = ctx_height (ctx);

  // we could keep track of when this is dirty and needs recomputation
  if (anim_t > 0.0f)
    overview_init (ctx);

  int rows = 1;
  int cols = 1;

  if (n_clients <= 4) { rows = 2; cols = 2;}

  while ( n_clients > cols * rows)
  {
    if (cols > rows * 1.2)
    {
       rows++;
    } else cols++;
  }

  float col_width  = screen_width / cols;
  float row_height = screen_height / rows;
  float icon_width = col_width - gap;
  //float icon_height = row_height - em;
  
  int i = 0;

  //ctx_rgb(ctx,0.1,0.1,0.1);
  //ctx_rgb (ctx, 0.086,0.086,0.113);
  cfg_ctx_color (&ctx_config, "style", "wallpaper", ctx);
  ctx_paint(ctx);

  for (i = 0; i < opos_count; i++)
  {
    CtxClient *client = opos[i].client;
    float x0 = opos[i].x0;
    float y0 = opos[i].y0;
    float x1 = opos[i].x1;
    float y1 = opos[i].y1;
    float w0 = opos[i].w0;
    float h0 = opos[i].h0;
    float scale0 = opos[i].scale0;
    float scale1 = opos[i].scale1;

    float x = x0 * (1.0-anim_t) + anim_t * x1;
    float y = y0 * (1.0-anim_t) + anim_t * y1;
    float scale = scale0 * (1.0-anim_t) + anim_t * scale1;
    float w = 0;
    float h;

    if (opos[i].id == ctx_clients_active (ctx))
    {
      w = w0 * (1.0-anim_t) + anim_t * icon_width;
      h = h0 * (w / w0);
    }
    else
    {
      w = icon_width;
      h = h0 * (w / w0);
    }

    ctx_save(ctx);
    ctx_scale (ctx, scale, scale);
    ctx_client_draw (ctx, client, x / scale, y / scale);
    ctx_restore(ctx);
    ctx_rectangle (ctx, x-2, y-2, w+4, h+4);
    ctx_listen (ctx, CTX_DRAG_PRESS, overview_select_client, client, NULL);
    ctx_reset_path (ctx);

    if (opos[i].id == ctx_clients_active (ctx))
    {
          if (anim_t > 0.95)
          {
            ctx_round_rectangle (ctx, x-2-0.5f, y-2-0.5f, w+4, h+4, 0.3 * em);
            cfg_ctx_color (&ctx_config, "style", "overview_fg_selected", ctx);



            ctx_line_width (ctx, cfg_cached_f32 (&ctx_config, "style", "overview_selected_line_width"));
            ctx_stroke(ctx);
            ctx_rgba(ctx,1,0,0,0.6);
            ctx_rectangle (ctx, x+w-em*2.5, y, em *2.5, em *2.5);
            ctx_listen (ctx, CTX_PRESS, remove_tab_cb, NULL, NULL);

            ctx_reset_path (ctx);
            icon_remove_tab (ctx, x+w-em*2.5, y, em *2.5, em *2.5);
            ctx_fill (ctx);

            float ty = y + row_height - em;
            if (y + h + 1.2 * em < ty)
               ty = y + h + 1.2 * em;
          
            const char *title = ctx_client_title (client);

            float title_width = ctx_text_width (ctx, title);

            ctx_round_rectangle (ctx, x+w/2 - title_width/2 - 0.5 * em,
                              ty - 1.0 * em, title_width + 1.0 * em,
                              1.4 * em, 1.4 * em);

            cfg_ctx_color (&ctx_config, "style", "overview_bg_shade", ctx);
            ctx_fill (ctx);

            ctx_move_to (ctx, x+w/2, ty);
  
            cfg_ctx_color (&ctx_config, "style", "overview_fg", ctx);

            ctx_text_align (ctx, CTX_TEXT_ALIGN_CENTER);
            ctx_text (ctx, ctx_client_title (client));
          }
    }
  }
  ctx_dock (ctx);
}


void tabs_and_overview (Ctx *ctx, float delta_s)
{
      int n_clients = ctx_list_length (ctx_clients (ctx));
          if (!force_full_size && (in_overview || leave_overview > 0.0f))
          {
            if (leave_overview > 0.0f)
            {
              leave_overview -= delta_s;
              if (leave_overview < 0.0f)
                leave_overview = 0.0f;
              overview (ctx, leave_overview / animation_duration);
              ctx_queue_draw (ctx);
            }
            else
            {
              leave_overview = 0.0f;
              overview_t += delta_s;
              if (overview_t >= animation_duration)
                overview_t = animation_duration;
              else
                ctx_queue_draw (ctx);
              overview (ctx, overview_t / animation_duration);
            }
          }
        else
          {
            clients_draw (ctx);
#if 1
            if ((n_clients != 1) || (ctx_clients (ctx) &&
                                 !flag_is_set(
                                         ctx_client_flags (((CtxClient*)ctx_clients(ctx)->data)), CTX_CLIENT_MAXIMIZED)))
              draw_panel (ctx);
#endif
            ctx_dock (ctx);
          }
}


#else

#if CTX_BIN_BUNDLE
int ctx_terminal_main (int argc, char **argv)
#else
int main (int argc, char **argv)
#endif
{
  return -1;
}

#endif
